#!/usr/bin/env python
# -*- coding: utf-8 -*-
import csv,datetime,logging,os,subprocess,sys,threading,time
import numpy as np
import random as rnd
from scipy.stats import  binom,expon,gamma,lognorm,norm,nbinom,poisson,uniform

class NGSAvailableDistributions:
	"""
	Class for the definition of the available distributions for this program.
	----
	Attributes:
	- relationNumParams: Dictionary with the pairing code:number_of_parameters.
	- distributions: Codes of all the available distributions.
	- names: Full name of the available distributions.
	"""

	relationNumParams=dict({\
		"b":2,#mean,percentages\
		"e":1,#mean\
		"f":1,\
		"g":2,\
		"g1":1,\
		"ln":2,\
		"n":2,\
		"nb":2,\
		"p":1,\
		"u":2\
		})
	distributions=relationNumParams.keys()
	names=dict({\
		"b":"Binomial",\
		"e":"Exponential",\
		"f":"Fixed",\
		"g":"Gamma",\
		"g1":"Gamma (mean=1)",\
		"ln":"Log Normal",\
		"n":"Normal",\
		"nb":"Negative Binomial",\
		"p":"Poisson",\
		"u":"Uniform"\
	})

class NGSPhyDistributionParser:
	"""
	Class for parsing the distribution "words" introduced in the settings file.
	----------------------------------------------------------------------------
	Attributes:
	- __params: to store distribution parameters
	- __name: to store distribution code name
	- __dependency: to identify whether the whole word should be taking into
	account or if the mean (or main) parameter of the distribution depends on
	previously calculated values.
	"""
	__params=[]
	__name=""
	__dependency=None

	def __init__(self,distroline,dependency):
		distroline=distroline.split(":")
		self.__dependency=dependency or False
		# i depend on another level value
		if self.__dependency:
			# Value is empty
			if (len(distroline)==1):
				# only have distribution name, meaning
				# distribution has only 1 param and it comes
				# from the upperLevelDistribution
				self.__name=distroline[0].lower()
			if (len(distroline)==2):
				# i have name and a parameter,
				# meaning, first param comes from above
				# second is the second parameter of my distro
				self.__name=distroline[0].lower()
				self.__params=distroline[1].split(",")
		else:
			self.__name=distroline[0].lower()
			if len(distroline) > 1:
				self.__params= distroline[1].split(",")


	def setParams(self,params):
		"""
		Allows to modify the values of the distribution parameters
		-----------------------------------------------------------------------
		- input: params
		"""
		self.__params=params

	def getParams(self):
		"""
		Returns distribution parameters
		"""
		return self.__params

	def printParams(self):
		"""
		Prints distribution parameters
		"""
		print(self.__params)

	def setName(self,name):
		"""
		Allows to modify the code name of the distribution
		-----------------------------------------------------------------------
		- input: name
		"""
		self.__name=name

	def getName(self):
		"""
		Returns distribution parameters
		"""
		return self.__name

	def printName(self):
		"""
		Prints distribution name
		"""
		print(self.__name)

	def asNGSPhyDistribution(self):
		"""
		Returns a NGSPhyDistribution object from the parsed distribution word.
		"""
		if not self.__dependency:
			return NGSPhyDistribution(self.__name,self.__params)
		else: return None

	def validate(self):
		"""
		Validates if the input of the distribution word correspond to its definition.
		Meaning, codename is within the available distribution code names,
		number of parameters correspond to the selected distribution and the
		parameters introduced are numbers.
		"""
		message="Coverage distribution parsing OK.\n\t"
		status=True
		if not self.__name in NGSAvailableDistributions.distributions:
			message="{0}\n\t{1}\n\t{2}".format(\
				"Problem parsing coverage distribution.",\
				"Distribution selected invalid."
				"Please verify. Exiting"\
				)
			status=False
		if (NGSAvailableDistributions.relationNumParams[self.__name]==len(self.__params)) or (NGSAvailableDistributions.relationNumParams[self.__name]-1==len(self.__params) and self.__dependency):
			# print(len(self.__params))
			for index in range(0,len(self.__params)):
				try:
					self.__params[index]=float(self.__params[index])
				except ValueError:
					message="{0}\n\t{1}\n\t{2}".format(\
					"Problem parsing coverage distribution.",\
					"Not a correct parameter value.",\
					"Please verify. Exiting"\
					)
					status=False
					break
		else:
			message="{0}\n\t{1}\n\t{2}\t{3} ({4}) - Num params: {5}\n\t{6}\t{7}\n\t{8}".format(\
				"Problem parsing coverage distribution.",\
				"Not the correct number of parameters for the selected distribution",\
				"Given: ",NGSAvailableDistributions.names[self.__name],self.__name,len(self.__params),\
				"Required: ",NGSAvailableDistributions.relationNumParams[self.__name],\
				"Please verify. Exiting"\
				)
			status=False

		return status, message

class NGSPhyDistribution:
	"""
	Class that samples values from the available distributions.
	----------------------------------------------------------------------------
	Attributes:
	- __name: Code name of the distribution
	- __params: Parameters of the distribution to be sampled.
	"""
	__name=None
	__params=[]
	def __init__(self,name,params):
		self.__name=name
		self.__params=params
		# print(self.__name)
		# print(self.__params)

	def value(self,samples=1):
		"""
		Samples number of values given from the specific distribution.
		------------------------------------------------------------------------
		- samples: number of values that will be returned.
		"""
		value=0
		try:
			for item in self.__params:
				if item==0: break
			if item==0: value=[0]*samples
			else:
				if self.__name=="b": value=self.binom(samples)
				if self.__name=="e": value=self.exponential(samples)
				if self.__name=="f": value=self.fixed(samples)
				if self.__name=="g": value=self.gamma(samples)
				if self.__name=="g1": value=self.gamma1(samples)
				if self.__name=="ln": value=self.lognormal(samples)
				if self.__name=="n": value=self.normal(samples)
				if self.__name=="nb": value=self.nbinom(samples)
				if self.__name=="p": value=self.poisson(samples)
				if self.__name=="u": value=self.uniform(samples)
		except Exception as ex:
			exc_type, exc_obj, exc_tb = sys.exc_info()
			fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
			message="\n\tUnexpected: {0} | {1} - File: {2} - Line:{3}".format(\
				ex,exc_type, fname, exc_tb.tb_lineno)
			status=False
			raise Exception(message)
			# self.appLogger.error(message)
			# sys.exit()

		return value

	def binom(self,samples):
		"""
		Sampling from a binomial distribution
		------------------------------------------------------------------------
		- samples: number of values that will be returned.
		"""
		#n=number of times tested
		#p=probability
		n=self.__params[0]
		p=self.__params[1]
		distro=binom(n,p)
		f=distro.rvs(size=samples)
		return f

	def exponential(self,samples):
		"""
		Sampling from a exponential distribution
		------------------------------------------------------------------------
		- samples: number of values that will be returned.
		"""
		f=np.random.exponential(float(self.__params[0]),size=samples)
		return value

	def fixed(self,samples):
		"""
		Fixed values
		------------------------------------------------------------------------
		- samples: number of values that will be returned.
		"""
		value=[self.__params[0]]*samples
		return value

	def gamma1(self,samples):
		"""
		Sampling from a Gamma distribution with mean 1
		------------------------------------------------------------------------
		- samples: number of values that will be returned.
		- gamma parameterized with shape and scale
		"""
		# E|x| = k.theta (alpha*theta)
		# If i want  mean=1, theta=E|x|/alpha=1/alpha
		shape=float(self.__params[0]*1.0)
		theta=float(1/(shape*1.0))
		distro=gamma(shape,theta)
		f=distro.rvs(size=samples)
		return f


	def gamma(self,samples):
		"""
		Sampling from a Gamma distribution with mean 1

		The parameterization with alpha and beta is more common in Bayesian statistics,
		where the gamma distribution is used as a conjugate prior distribution
		for various types of inverse scale (aka rate) parameters, such as the
		lambda of an exponential distribution or a Poisson distribution[4]  or for
		t hat matter, the beta of the gamma distribution itself.
		(The closely related inverse gamma distribution is used as a conjugate
		prior for scale parameters, such as the variance of a normal distribution.)
		shape, scale = 2., 2. # mean=4, std=2*sqrt(2)

		(Wikipedia: https://en.wikipedia.org/wiki/Gamma_distribution)

		s = np.random.gamma(shape, scale, 1000)
		E|x| = k.theta (alpha*theta)
		If i want a specific mean, theta=E|x|/alpha
		------------------------------------------------------------------------

		- samples: number of values that will be returned.
		"""
		shape=float(self.__params[0]*1.0)
		theta=float(self.__params[1]*1.0)
		distro=gamma(shape,theta)
		f=distro.rvs(size=samples)
		return f


	def lognormal(self,samples):
		"""
		Sampling from a Log Normal distribution

		Parameters:
		m=mean
		s=sigma - standard deviation
		------------------------------------------------------------------------
		- samples: number of values that will be returned.
		"""
		mean=float(self.__params[0]*1.0)
		sigma=float(self.__params[1]*1.0)
		distro=lognorm(s=sigma,scale=np.exp(mean))
		f=distro.rvs(size=samples)
		return f

	def nbinom(self,samples):
		"""
		Sampling from a Negative binomial distribution

		Parameters:
		mu= poissonon mean
		r controls the deviation from the poisson

		This makes the negative binomial distribution suitable as a robust
		alternative to the Poisson, which approaches the Poisson for large r,
		but which has larger variance than the Poisson for small r.
		------------------------------------------------------------------------
		- samples: number of values that will be returned.
		"""
		mu=float(self.__params[0])
		r=float(self.__params[1])
		p=(r*1.0)/(r+mu)
		distro=nbinom(r,p)
		f=distro.rvs(size=samples)
		return f

	def normal(self,samples):
		"""
		Sampling from a Normal distribution

		Parameters:
		mean (location)
		variance (squared scale)
		------------------------------------------------------------------------
		- samples: number of values that will be returned.
		"""
		locMean=float(self.__params[0]*1.0)
		scaleVariance=np.sqrt(float(self.__params[1]*1.0))
		distro=norm(loc=locMean,scale=scaleVariance)
		f=distro.rvs(size=samples)
		return f

	def poisson(self,samples):
		"""
		Sampling from a Poisson distribution
		Parameters:
		mean
		------------------------------------------------------------------------
		- samples: number of values that will be returned.
		"""
		l=float(self.__params[0]*1.0)
		distro=poisson(l)
		f=distro.rvs(size=samples)
		return f

	def uniform(self,samples):
		"""
		Sampling from a Poisson distribution
		Parameters:
		meean
		------------------------------------------------------------------------
		- samples: number of values that will be returned.
		"""
		minParam=float(self.__params[0]*1.0)
		maxParam=float(self.__params[1]*1.0)
		f= np.random.uniform(low=minParam,high=maxParam,size=samples)
		return f

class CoverageMatrixGenerator:
	"""
	Class for the calculation of coverage matrices.
	Able to intoduced:
	 	- "genomic noise" for entire datasets.
		- Variation across individuals and loci.
		- on/off target effect
		- coverage decay related to distance to the reference used
	----------------------------------------------------------------------------
	Attributes:
	- OFFTARGET_COVERAGE_FRACTION: Specific for the on/off target effect.
	- appLogger: Logger object
	- settings: Settings object withh all the program parameters
	- alphashapesLocus: parameters (alpha shapes) for the Gamma distribution with
	mean 1 that will be used as multiplier for across loci coverage variation.
	- alphashapesIndividuals: parameters (alpha shapes) for the Gamma distribution
	with mean 1 that will be used as multiplier for across individuals coverage
	variation.
	- locusMultiplier: parameters to be used to sample alphashapesLocus.
	- individualsMultiplier: parameters to be used to sample alphashapesIndividuals.
	- experiment: expected coverage for a specific replicate (species tree data).
	- outputFolderPath: folder where coverage matrices will be stored.
	- numReplicates: number of species trees.
	- numLociPerReplicate: number of loci per species tree.
	- numIndividualsPerReplicate: number of individuals per species tree.
	- numReplicateDigits: number of digits needed to represent numReplicates.
	- numLociPerReplicateDigits: number of digits needed to represent.
	- filteredReplicates: identifier of the species trees that will be used.

	"""
	# OFFTARGET_COVERAGE_FRACTION=0.1#
	appLogger=None
	settings=None

	alphashapesLocus=[]
	alphashapesIndividuals=[]
	locusMultiplier=[]
	individualsMultiplier=[]

	experiment=None
	experimentValues=None

	outputFolderPath=""

	numReplicates=0
	numLociPerReplicate=[]
	numIndividualsPerReplicate=[]
	numReplicateDigits=0
	numLociPerReplicateDigits=[]
	filteredReplicates=[]

	onTargetLoci=[]
	offTargetLoci=[]
	notcaptured=[]
	numOnTargetLoci=0

	def __init__(self, settings):
		self.appLogger=logging.getLogger("ngsphy")
		self.appLogger.debug("Init class Coverage Matrix Generator...")
		self.settings=settings
		self.experiment=self.settings.experiment.asNGSPhyDistribution()
		self.numReplicates=self.settings.numReplicates
		cc=self.settings.parser.get("general", "numlociperreplicate").strip().split(",")
		self.numLociPerReplicate=[ int(item) for item in cc if not item ==""]
		cc=self.settings.parser.get("general", "numindividualsperreplicate").strip().split(",")
		self.numIndividualsPerReplicate=[ int(item) for item in cc if not item == ""]
		self.numReplicateDigits=len(str(self.numReplicates))
		self.numLociPerReplicateDigits=[len(str(item)) for item in self.numLociPerReplicate]
		self.numIndividualsPerReplicateDigits=[len(str(item )) for item in self.numIndividualsPerReplicate]
		self.filteredReplicates=[int(item) for item in self.settings.parser.get("general", "filtered_replicates").strip().split(",")]
		self.experimentValues=[1]*len(self.filteredReplicates)
		if (self.settings.locus):
			distro=self.settings.locus.asNGSPhyDistribution()
			self.alphashapesLocus=distro.value(len(self.filteredReplicates))
			self.locusMultiplier=[NGSPhyDistribution("g1",[self.alphashapesLocus[index]]).value(self.numLociPerReplicate[index]) for index in range(0,len(self.filteredReplicates))]
		if (self.settings.individual):
			distro=self.settings.individual.asNGSPhyDistribution()
			self.alphashapesIndividuals=distro.value(len(self.filteredReplicates))
			self.individualsMultiplier=[NGSPhyDistribution("g1",[self.alphashapesIndividuals[index]]).value(self.numIndividualsPerReplicate[index]) for index in range(0,len(self.filteredReplicates))]
		self.generateFolderStructure()

	def generateFolderStructure(self):
		"""
		Generation of folder structure needed for the coverage matrices.
		"""
		self.appLogger.debug("Coverage folder structure generation...")
		try:
			os.makedirs(self.settings.coverageFolderPath)
			self.appLogger.info("Generating output folder ({0})".format(self.settings.coverageFolderPath))
		except:
			self.appLogger.debug("Output folder exists ({0})".format(self.settings.coverageFolderPath))

	def calculate(self):
		"""
		Calulation of coverage matrices
		"""
		message=""
		status=True;
		# for indexRep in self.filteredReplicates:
		self.appLogger.info("Coverage calulation: ReplicateID - currentReplicate/numberOfWorkingReplicates (numIndividuals,numLoci)")
		for index in range(0,len(self.filteredReplicates)):
			nInds=self.numIndividualsPerReplicate[index]
			nLoci=self.numLociPerReplicate[index]
			# expectedCoverage=self.experiment.value(nInds*nLoci)
			val=self.experiment.value(1)
			self.experimentValues[index]=val
			# self.appLogger.debug(val)
			expectedCoverage=[val]*(nInds*nLoci)
			coverageMatrix=np.matrix(expectedCoverage)
			coverageMatrix.shape=[nInds,nLoci]
			# individuals + loci coverage variation
			# individuals + loci multipliers
			self.appLogger.info("ReplicateID {0} - {1}/{2} ({3},{4})".format(\
				self.filteredReplicates[index],\
				index+1,\
				len(self.filteredReplicates),\
				nInds,\
				nLoci))
			if self.settings.locus:
				try:
					self.appLogger.debug("Locus-wide multipliers")
					multipliers=self.locusMultiplier[index]
					for loc in range(0,nLoci):
						coverageMatrix[:, loc]=coverageMatrix[:,loc]*multipliers[loc]
				except Exception as ex:
					exc_type, exc_obj, exc_tb = sys.exc_info()
					fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
					message="{0}\n\t{1}\n\t{2} | File: {3} - Line:{4}".format(\
						"Unexpected. Coverage computation. Locus-wide multipliers.",
						ex,exc_type, fname, exc_tb.tb_lineno)
					return False, message
			if self.settings.individual:
				try:
					self.appLogger.debug("individual-wide multipliers")
					multipliers=self.individualsMultiplier[index]
					for ind in range(0,nInds):
						coverageMatrix[ind,:]=coverageMatrix[ind,:]*multipliers[ind]
				except Exception as ex:
					exc_type, exc_obj, exc_tb = sys.exc_info()
					fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
					message="{0}\n\t{1}\n\t{2} | File: {3} - Line:{4}".format(\
						"Unexpected. Coverage computation. Individual-wide multipliers.",
						ex,exc_type, fname, exc_tb.tb_lineno)
					return False, message
			# taxonomic variation
			# I need the file with the realtion of the individuals
			# dont car if haploid or diploid because ill only need the first 3 columnsg
			# which match in both cases
			if self.settings.taxon:
				self.appLogger.debug("Taxon-specific multipliers")
				individualsTableFilename=os.path.join(\
					self.settings.tablesFolderPath,
					"{0}.{1:0{2}d}.individuals.csv".format(\
						self.settings.projectName,\
						self.filteredReplicates[index],\
						self.numReplicateDigits)
					)
				individualsTable=None
				with open(individualsTableFilename) as individualsTableFile:
					d = csv.DictReader(individualsTableFile)
					individualsTable = [row for row in d]

				for row in individualsTable:
					if str(row['spID']) in self.settings.taxon.keys():
						coverageMatrix[ind,]=np.array(coverageMatrix[ind,])*self.settings.taxon[str(row['spID'])]
			#on/off target
			self.onTargetLoci=range(0,nLoci)
			self.offTargetLoci=[]
			try:
				if self.settings.offtarget["loci"] > 0:
					self.appLogger.debug("Off-target coverage")
					nSamples=int(self.settings.offtarget["loci"]*nLoci)
					self.offTargetLoci=np.random.choice(nLoci,nSamples,replace=False)
					self.onTargetLoci=set(range(0,nLoci))-set(self.offTargetLoci)
					for loc in self.offTargetLoci:
						coverageMatrix[:,loc]=coverageMatrix[:,loc]*self.settings.offtarget["coverage"]
			except Exception as ex:
				return False, "\n\t{0}\n\t{1}\n\t{2}\n\t{3}".format(\
					ex,\
					"Coverage Matrix Generator",\
					"An unexpected error occurr with the on/off target parameter. ",\
					"Please verify. Exiting.")

			# not captured
			self.numOnTargetLoci=nLoci-len(self.offTargetLoci)
			try:
				if self.settings.notcaptured:
					# random number of loci  (respecting probabilities)
					self.appLogger.debug("Not captured on target")
					nSamples=int(self.settings.notcaptured*self.numOnTargetLoci)
					self.notcaptured=np.random.choice(self.numOnTargetLoci,nSamples,replace=False)
					for loc in self.notcaptured:
						coverageMatrix[:,loc]=coverageMatrix[:,loc]*0
			except Exception as ex:
				return False, "\n\t{0}\n\t{1}\n\t{2}\n\t{3}".format(\
					ex,\
					"Coverage Matrix Generator",\
					"An unexpected error occurr with the not captured parameter. ",\
					"Please verify. Exiting.")

			# print(coverageMatrix)
			self.write(coverageMatrix,self.filteredReplicates[index])
		return status,message

	def write(self,coverageMatrix,indexRep):
		"""
		Writing into file the coverage matrix
		"""
		index=self.filteredReplicates.index(indexRep)
		self.appLogger.debug(\
			"Writing coverage matrix for replicate: {0:0{1}d}".format(\
				self.filteredReplicates[index],\
				self.numReplicateDigits))
		filename=os.path.join(\
			self.settings.coverageFolderPath,\
			"{0}.{1:0{2}d}.coverage.csv".format(\
				self.settings.projectName,\
				self.filteredReplicates[index],\
				self.numReplicateDigits\
			)
		)
		filepath=os.path.abspath(filename)
		header= ["indID"] + \
				[ "L.{0:0{1}d}".format(\
					loc+1,\
					self.numLociPerReplicateDigits[index])\
				for loc in range(0,self.numLociPerReplicate[index])]
		nInds=self.numIndividualsPerReplicate[index]
		indIds=np.transpose(np.array(range(0,nInds)))
		f=open(filepath,"w")
		f.write("{}\n".format(" ,".join(header)))
		for r in range(0,nInds):
			weird=coverageMatrix[r,].tolist()[0]
			weird=["{0:.3f}".format(float(item)) for item in weird]
			line=", ".join(weird)
			line=line.replace("[","").replace("]","")
			# print(line)
			f.write("{0},{1}\n".format(r,line))
		f.close()

		filename=os.path.join(\
			self.settings.coverageFolderPath,\
			"{0}.{1:0{2}d}.multipliers.txt".format(\
				self.settings.projectName,\
				self.filteredReplicates[index],\
				self.numReplicateDigits\
			)
		)
		filepath=os.path.abspath(filename)
		f=open(filepath,"w")
		f.write("# Experiment coverage:\n")
		f.write("{0}\n".format(self.experimentValues[index]))
		f.write("# Locus alpha shapes:\n")
		if not len(self.alphashapesLocus) == 0:
			f.write("{0}\n".format(self.alphashapesLocus[index]))
		else:
			f.write("None\n")
		f.write("# Individual alpha shapes:\n")
		if not len(self.alphashapesIndividuals)==0:
			f.write("{0}\n".format(self.alphashapesIndividuals[index]))
		else:
			f.write("None\n")

		f.write("# Locus Multipliers:\n")
		if not len(self.locusMultiplier) ==0:
			for item in range(0, len(self.locusMultiplier[index])):
				f.write("{0}\t{1}\n".format(item+1, self.locusMultiplier[index][item]))
		else:
			f.write("None\n")
		f.write("# Individual Multipliers:\n")
		if not len(self.individualsMultiplier) ==0:
			for item in range(0, len(self.individualsMultiplier[index])):
				f.write("{0}\t{1}\n".format(item+1, self.individualsMultiplier[index][item]))
		else:
			f.write("None\n")
		f.write("# On target loci\n")
		f.write(",".join([str(item+1) for item in self.onTargetLoci]))
		f.write("\n# Off target loci\n")
		f.write(",".join([str(item+1) for item in self.offTargetLoci]))
		f.write("\n# Not captured loci\n")
		f.write(",".join([str(item+1) for item in self.notcaptured]))
		f.write("\n# Taxon decay: \n")
		for item in self.settings.taxon.keys():
			f.write("{0},{1}\n".format(item, self.settings.taxon[item]))
		f.close()



class NGSphyException(Exception):
    """
    Exception raised for errors of the NGSphy program.
    ----------------------------------------------------------------------------
    Attributes:
        expression -- input expression in which the error occurred
        message -- explanation of the error
    """

    def __init__(self, expression, message, time):
        self.expression = expression
        self.message = message
        self.time= time

class NGSphyExitException(Exception):
    """
    Exception raised for ending of the NGSphy process.
    ----------------------------------------------------------------------------
    Attributes:
        expression -- input expression in which the error occurred
        message -- explanation of the error
    """
    def __init__(self, expression, message, time):
        self.expression = expression
        self.message = message
        self.time= time



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import dendropy,logging,os,shutil,sys
import settings as sp

class Rerooter:
	"""
	Class used for the manipulation of gene trees, specifically, to handle
	the re-rooting and prunning of the given gene tree.
	----------------------------------------------------------------------------
	Attributes:
	- appLogger: logger to store status of the process flow
	- settings: Settings object withh all the program parameters
	- tree: final gene tree
	- outputFilename: filename where the new tree will be stored
	- outputFilePath: absolute path of the filename where the new tree will be stored
	"""
	# general
	appLoger=None
	settings=None
	# tree-related
	tree=None
	outputFilename="ngsphy.tree"
	outputFilePath=""

	def __init__(self, settings):
		self.appLogger=logging.getLogger("ngsphy")
		self.appLogger.debug('Rerooting')
		self.settings=settings
		self.outputFilePath=os.path.join(\
			self.settings.basepath,\
			"1",\
			self.outputFilename\
		)
		self.generateFolderStructure()

	def generateFolderStructure(self):
		"""
		Generation of basic folder structure for this process
		"""
		folder=os.path.join(self.settings.basepath,"1")
		try:
			os.makedirs(folder)
			self.appLogger.info("Generating alignments folder ({0})".format(folder))
		except:
			self.appLogger.debug("Project folder exists ({0})".format(folder))
		try:
			shutil.copyfile(self.settings.ancestralSequenceFilePath, os.path.join(folder, "ancestral.fasta"))
		except:
			self.appLogger.debug("File already exists in this location ({0})".format(os.path.join(folder, "ancestral.fasta")))


	def run(self):
		"""
		Process flow of the re-rooting and pruninc process
		"""
		self.appLogger.debug('Running rerooting')
		try:
			self.tree=dendropy.Tree.get(path=self.settings.geneTreeFile,\
			 	schema="newick",\
				rooting='force-rooted',\
				preserve_underscores=True)
		except Exception as ex:
			return False, ex
		newroot=self.tree.find_node_with_taxon_label(self.settings.anchorTipLabel)
		lennewroot=newroot.edge_length
		if newroot:
			self.tree.reroot_at_edge(newroot.edge,length1=lennewroot,length2=0, update_bipartitions=False)
			self.settings.geneTreeFile=self.outputFilePath
		else:
			return False, "{0}\n\t{1}\n\t{2}".format(\
			"Rerooting problem.",\
			"Something might be wrong with the reference label.",\
			"Please Verify. Exiting"\
			)
		self.writeTreeIntoFile()
		return True,""

	def writeTreeIntoFile(self):
		"""
		Writes into a file the resulting tree
		"""
		self.appLogger.debug("Writing tree")
		self.appLogger.debug("Writing tree")
		self.tree.write(\
			path=self.outputFilePath,\
			schema="newick",\
			suppress_rooting=True\
			)

	def recheckPloidyAfterRerooting(self):
		status=True;message=""
		try:
			self.appLogger.debug("Ploidy after reroot")
			leaves = [node for node in self.tree if node.taxon]
			# print(self.tree.print_plot())
			leaves = [node.taxon.label for node in leaves if not node.taxon.label==self.settings.anchorTipLabel]
			# print(self.tree.as_string(schema='newick'))
			labels=dict()
			for tip in leaves:
				tmp=tip.strip().split("_")
				geneFamily="{0}_{1}".format(tmp[0],tmp[1])
				try:
					f=labels[geneFamily]
					labels[geneFamily]+=1
				except:
					labels[geneFamily]=1
			fail=False
			for item in labels.keys():
				if not labels[item] % self.settings.ploidy == 0:
					fail=True
					break
			if fail:
				status=False
				return status, "\n\t{0}\n\t{1}\n\t{2}".format(\
					"Problems for the individual assignment.",\
					"Number of gene-tree tips per species is not valid for the given ploidy.",\
					"Please verify. Exiting."\
				)
		except Exception as ex:
			exc_type, exc_obj, exc_tb = sys.exc_info()
			fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
			message="\n\t{0} {1} | {2} - File: {3} - Line:{4}\n\t{5}".format(\
			    "Rerooter: Something is wrong.",\
			    ex,exc_type,\
				fname, exc_tb.tb_lineno,\
			    "Please verify. Exiting."
			    )
			status=False
		return status, message



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import logging, platform

LOG_LEVEL_CHOICES=["DEBUG","INFO","WARNING","ERROR"]
class MEOutputFormatter:
    """
    This module incorporates different constant color values from bash shell
    for a "pretty" visualization of the log.
    """
    PURPLE = '\033[95m'
    CYAN = '\033[96m'
    DARKCYAN = '\033[36m'
    BLUE = '\033[94m'
    GREEN = '\033[92m'
    YELLOW = '\033[93m'
    RED = '\033[91m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'
    END = '\033[0m'

class MELoggingFormatter(logging.Formatter):
    """
    This module has been done to modify the way in which the log is written to
    the standard output and the log file.
    """

    FORMATS={}

    if platform.system()=="Darwin":
        FORMATS={\
            "CONFIG":"%(asctime)s - {0}{1}%(levelname)s{2}:\t%(message)s".format(\
                MEOutputFormatter.BOLD, MEOutputFormatter.BLUE,MEOutputFormatter.END),\
            "ERROR": "%(asctime)s - {0}{1}%(levelname)s (%(module)s){2}{0}:\t%(message)s{2}".format(\
                MEOutputFormatter.BOLD, MEOutputFormatter.RED,MEOutputFormatter.END),\
            "WARNING":"%(asctime)s - {0}{1}%(levelname)s{2}:\t%(message)s".format(\
                MEOutputFormatter.BOLD, MEOutputFormatter.YELLOW,MEOutputFormatter.END),\
            "INFO": "%(asctime)s - {0}{1}%(levelname)s{2}:\t%(message)s".format(\
                MEOutputFormatter.BOLD, MEOutputFormatter.GREEN,MEOutputFormatter.END),\
            "DEBUG":"%(asctime)s - {0}{1}%(levelname)s{2} (%(module)s:%(lineno)d):\t%(message)s".format(\
                MEOutputFormatter.BOLD, MEOutputFormatter.PURPLE,MEOutputFormatter.END),\
            "DEFAULT":"%(asctime)s - {0}%(levelname)s{1}:\t%(message)s".format(\
                MEOutputFormatter.BOLD,MEOutputFormatter.END)\
           }
    else:
        FORMATS={\
            "CONFIG":"%(asctime)s - {0}{1}%(levelname)s{2}:\t%(message)s".format(\
                MEOutputFormatter.BOLD, MEOutputFormatter.BLUE,MEOutputFormatter.END),\
            "ERROR": "%(asctime)s - {0}{1}%(levelname)s (%(module)s){2}{0}:\t%(message)s{2}".format(\
                MEOutputFormatter.BOLD, MEOutputFormatter.RED,MEOutputFormatter.END),\
            "WARNING":"%(asctime)s - {0}{1}%(levelname)s{2}:\t%(message)s".format(\
                MEOutputFormatter.BOLD, MEOutputFormatter.YELLOW,MEOutputFormatter.END),\
            "INFO": "%(asctime)s - {0}{1}%(levelname)s{2}:\t%(message)s".format(\
                MEOutputFormatter.BOLD, MEOutputFormatter.GREEN,MEOutputFormatter.END),\
            "DEBUG":"%(asctime)s - {0}{1}%(levelname)s{2} (%(module)s|%(funcName)s:%(lineno)d):\t%(message)s".format(\
                MEOutputFormatter.BOLD, MEOutputFormatter.PURPLE,MEOutputFormatter.END),\
            "DEFAULT":"%(asctime)s - {0}%(levelname)s{1}:\t%(message)s".format(\
                MEOutputFormatter.BOLD,MEOutputFormatter.END)\
           }

    def __init__(self,fmt,datefmt):
        logging.Formatter.__init__(self, fmt, datefmt)

    def format(self, record):
        original_fmt=self._fmt
        try:
            self._fmt = MELoggingFormatter.FORMATS[record.levelname]
        except:
            self._fmt = MELoggingFormatter.FORMATS["DEFAULT"]

        return logging.Formatter.format(self, record)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import argparse,datetime,logging,os,threading,sys
import numpy as np
import random as rnd
import individual as ig
import coverage
import sequence as sg
import settings as sp
import reads as ngs
import readcounts as rc
import rerooter as rr
from select import select
from ngsphyexceptions import *

class NGSphy:
    """
    Class that structures the whole process flow of the NGSphy: a genome-wide
    simulation of next-generation sequencing data from genetrees and species tree
    distributions.
    ----------------------------------------------------------------------------
    Attributes:
	- appLogger: Logger object to store status of the process flow.
	- path: current working directory
	- startTime: timing when processing has started
	- endTime: timing when processing has finished
	- settingsFile: path of the settings file.

    """
    appLogger=None
    path=os.getcwd()
    startTime=None
    endTime=None
    settingsFile=""

    def __init__(self,args):
    	self.startTime=datetime.datetime.now()
    	self.appLogger=logging.getLogger("ngsphy")
    	self.appLogger.info("Starting")
    	self.settingsFile=""
    	if (args.settings):
    		self.settingsFile=os.path.abspath(args.settings)
    	else:
    		self.settingsFile=os.path.abspath(os.path.join(os.getcwd(),"settings.txt"))
        if not (os.path.exists(self.settingsFile)):
            self.endTime=datetime.datetime.now()
            message="\t{0}\n\t{1}\n\t{2}\n\t{3}\n".format(\
                "The settings file:",\
                self.settingsFile,\
                "Does not exist...",\
                "Please verify. Exiting."
            )
            raise NGSphyExitException(ex.expression,message,self.endTime)

    def run(self):
        """
        Execution of the process flow.
        It does not return values. Will exit any time a process is wrongly
        executed or conditions of execution are not satisfied.
        """
    	# checking existence of settings file
    	try:
            status=True
            message="NGSphy finished correctly."
            self.appLogger.info("Checking settings...")
            if (os.path.exists(self.settingsFile)):
                self.appLogger.debug("Starting process")
                self.appLogger.debug("Starting process")
                # settings file exist, go ahead and run
            	self.settings=sp.Settings(self.settingsFile)
            	settingsOk,settingsMessage=self.settings.checkArgs()
            	# Exit heere if not correnct parser of the settings
            	if not settingsOk: self.ending(settingsOk,settingsMessage)
            	# Generate BASIC folder structure - output folder
            	self.generateFolderStructure()

            	# Settings exist and are ok.
            	# reroot-tree
                if self.settings.inputmode == 3:
                    self.appLogger.info("Re-rooting")
                    self.rerooter=rr.Rerooter(self.settings)
                    status, message=self.rerooter.run()
                    if not status: self.ending(status,message)
                    status, message= self.rerooter.recheckPloidyAfterRerooting()
                    if not status: self.ending(status, message)
                # Generate sequences
                if self.settings.inputmode < 4:
                    self.appLogger.info("Running sequence generator")
                    self.seqGenerator=sg.SequenceGenerator(self.settings)
                    # only copying ancestral sequences if the mode needs it
                    if(self.settings.inputmode >1):
                        statusRefSeq,messageRefSeq=self.seqGenerator.copyAncestralSequenceToOutputFolder()
                        if not statusRefSeq: self.ending(statusRefSeq,messageRefSeq)
                    indelibleStatus,indelibleMessage=self.seqGenerator.run()
                    if not (indelibleStatus): self.ending(indelibleStatus, indelibleMessage)
                # Generate Individuals (plody independency)
                self.indGenerator=ig.IndividualAssignment(self.settings)
                matingOk,matingMessage=self.indGenerator.checkArgs()
                if not matingOk: self.ending(matingOk,matingMessage)
                indelsCheckOK, indelsMessage=self.indGenerator.checkFilesForIndels()
                if not indelsCheckOK: self.ending(indelsCheckOK, indelsMessage)
                self.settings.indels=self.indGenerator.indelsPresence
                self.indGenerator.iteratingOverReplicates()
                # now checking NGS MODE
                if self.settings.ngsmode>0:
                    # Have to calculate coverage
                	self.appLogger.debug("Checking for Coverage. ")
                	covGenerator=coverage.CoverageMatrixGenerator(self.settings)
                	status,message=covGenerator.calculate()
                	if not status: self.ending(status,message)
            	# After this I'll have generated the individuals and folder structure
            	if self.settings.ngsmode==1:
            		self.appLogger.info("NGS Illumina reads - ART mode")
            		self.ngs=ngs.ARTIllumina(self.settings)
            		status, message=self.ngs.run()
            		if not status: self.ending(status,message)
            		self.appLogger.info("NGS read simulation process finished.")
            	elif self.settings.ngsmode==2:
            		self.appLogger.info("Read counts mode")
            		# self.appLogger.info("NGS read simulation is not being made.")
            		# If i have read count folder structure must change - i need reference folder
            		# print("ngsphy.py - Read count")
            		if self.settings.indels:
            			self.ending(False,"{0}\n\t{1}\n\t{2}".format(\
            				"Read Counts does not support INDELs (for now)",\
            				"Check the output folder. Data has been generated.",\
            				"Exiting."\
            			) )
            		self.readcount=rc.ReadCounts(self.settings)
            		status, message=self.readcount.run()
            		if not status: self.ending(status,message)
            	else:
            		self.appLogger.info("NGS simulation is not being made.")
            else:
            	self.ending(False,"Settings file ({0}) does not exist. Exiting. ".format(self.settingsFile))
            self.ending(status,message)
    	except NGSphyException as ex:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
            message="{1} | File: {2} - Line:{3}\n\t{0}".format(\
                ex.message,exc_type,\
                fname, exc_tb.tb_lineno\
            )
            raise NGSphyExitException(ex.expression,message, ex.time)

    def generateFolderStructure(self):
        """
        Generates basic folder structure for a NGSphy run.
        """
    	self.appLogger.info("Creating basic folder structure.")
    	try:
    		self.appLogger.info("Creating output folder: {0} ".format(self.settings.outputFolderPath))
    		os.makedirs(self.settings.outputFolderPath)
    	except:
    		self.appLogger.debug("Output folder ({0}) exists. ".format(self.settings.outputFolderPath))


    def generateSummaryLog(self, time):
    	filename=os.path.join(\
    		self.settings.outputFolderPath,\
    		"{0}.{1:%Y}{1:%m}{1:%d}-{1:%H}:{1:%M}:{1:%S}.summary.log".format(\
    			"NGSPHY",\
    			time\
    		)
    	)
    	with open(filename, 'w') as f:
            f.write(self.settings.formatSettingsMessage())

    def ending(self, good, message):
        """
        Handles the ending of the processes. Since there is some dependency
        between processes, this will also terminate the execution of the program
        and so, ending time must be given and also notified the status of the
        termination.
        ------------------------------------------------------------------------
        Parameters:
        - good: indicates whether the process terminated correctly.
        - message: Message indicating why process is terminated.
        Raises a NGSphyException
        """
    	self.endTime=datetime.datetime.now()
        if (os.path.exists(self.settings.outputFolderPath)):
            self.generateSummaryLog(self.endTime)
        eta=self.endTime-self.startTime
        if good:
            raise NGSphyExitException(good,message,eta)
        else:
            raise NGSphyException(good,message,eta)
    	# sys.exit()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import numpy as np
def parseMSAFile(fastapath):
	"""
	Parses a FASTA multiple sequence alignment file into a dictionary.
	---------------------------------------------------------------------------
	Parameters:
	- fastapath: full path of the FASTA MSA file.
	Returns:
	- A dictionary, where:
		- The keys of the dicitionary depend on the description of the sequences.
		Assuming the format of the description follows the
		<SpeciesTreeID>_<LocusTreeID>_<GeneTreeID> convention. The key will then
		correspond to the gene family of the sequence <SpeciesTreeID>_<LocusTreeID>.
		- Each element under a gene family, will be another dictionary which key is
		the <GeneTreeID> and the elements are the full description and the sequence
		itself.
	"""
	seqdata=[]
	seqDict=dict()
	description=""; seq=""; tmp="";count=1
	with open(fastapath,"r") as fastafile:
		for line in fastafile:
			if not (line.strip()==''):
				if line.strip().startswith(">"):
					seqdata+=[line.strip()]
				else:
					if(seqdata[-1].startswith(">")):
						seqdata+=[line.strip()]
					else:
						seqdata[-1]+=line.strip()

	for line in seqdata:
		if (count%2==0):
			seq=line
			try:
				test=seqDict[tag]
			except:
				seqDict[tag]={}

			try:
				seqDict[tag].update({tmp[2]:{\
					'description':description,\
					'sequence':seq\
				}})
			except:
				seqDict[tag][tmp[2]]={}
				seqDict[tag].update({tmp[2]:{\
					'description':description,\
					'sequence':seq\
				}})
			seq=None
			description=None
			tmp=None
		else:
			description=line[1:len(line)]
			tmp=description.split("_")
			tag="{0}_{1}".format(tmp[0], tmp[1])
		count+=1
	return seqDict


def parseMSAFileWithDescriptions(fastapath):
	"""
	Parses a FASTA multiple sequence alignment file into a dictionary.
	---------------------------------------------------------------------------
	Parameters:
	- fastapath: full path of the FASTA MSA file.
	Returns:
	- A dictionary, where:
		- The keys of the dicitionary are the description of the sequences.
		- Each element will contain the correspondin sequences.
	"""
	seqDict=dict()
	description=""; seq=""; tmp="";count=1
	seqdata=[]
	with open(fastapath,"r") as fastafile:
		for line in fastafile:
			if not (line.strip()==''):
				if line.strip().startswith(">"):
					seqdata+=[line.strip()]
				else:
					if(seqdata[-1].startswith(">")):
						seqdata+=[line.strip()]
					else:
						seqdata[-1]+=line.strip()
	for line in seqdata:
		if (count%2==0):
			seq=line.strip()
			seqDict[tmp]=seq
			seq=None
			description=None
			tmp=None
		else:
			description=line.strip()
			tmp=description[1:len(description)]
		count+=1
	return seqDict

def isFasta(filepath):
	"""
	Checks whether the format of the filepath introduced corresponds to FASTA
	---------------------------------------------------------------------------
	Parameters:
	- fastapath: full path of the FASTA MSA file.
	Returns:
	- A bollean. TRUE if it is a proper fasta, FALSE otherwise.
	"""
	fastaOk=True
	seqdata=[]
	seqDict=dict()
	description=""; seq=""; tmp="";count=1
	with open(filepath,"r") as fastafile:
		for line in fastafile:
			if not (line.strip()==''):
				if line.strip().startswith(">"):
					seqdata+=[line.strip()]
				else:
					if(seqdata[-1].startswith(">")):
						seqdata+=[line.strip()]
					else:
						seqdata[-1]+=line.strip()

	numLinesFile=len(seqdata)
	if (numLinesFile % 2 == 0): # It looks like a fasta, has even num of lines
		numSeqs=len(seqdata)/2
		index=0; properFormat=True
		while ((index < numSeqs) and properFormat):
			if (seqdata[index*2].startswith("+")):
				properFormat=False
			index=index+1
		if not index >= numSeqs: # there must have been an error
			return False
		if not properFormat:
			return False
	else:
		return False
	return True



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import argparse,datetime,logging, ngsphy, os,sys, platform
import loggingformatter as lf
################################################################################
# CONSTANTS
VERSION=1
MIN_VERSION=0
FIX_VERSION=11
PROGRAM_NAME="ngsphy.py"
AUTHOR="Merly Escalona <merlyescalona@uvigo.es>"
INSTITUTION="University of Vigo, Spain."
LOG_LEVEL_CHOICES=["DEBUG","INFO","WARNING","ERROR"]
LINE="--------------------------------------------------------------------------------"
################################################################################
# Logger initialization
ch = logging.StreamHandler()
loggerFormatter=lf.MELoggingFormatter(\
	fmt="%(asctime)s - %(levelname)s:\t%(message)s",\
	datefmt="%d/%m/%Y %I:%M:%S %p")
ch.setFormatter(loggerFormatter)
ch.setLevel(logging.NOTSET)
APPLOGGER=logging.getLogger("ngsphy")
APPLOGGER.addHandler(ch)

################################################################################
def createLogFile():
	formatString=""
	if platform.system()=="Darwin":
		formatString="%(asctime)s - %(levelname)s (%(module)s:%(lineno)d):\t%(message)s"
	else:
		formatString="%(asctime)s - %(levelname)s (%(module)s|%(funcName)s:%(lineno)d):\t%(message)s"
	logging.basicConfig(\
		filename=os.path.join(
				os.getcwd(),\
				"{0}.{1:%Y}{1:%m}{1:%d}-{1:%H}:{1:%M}:{1:%S}.log".format(\
					PROGRAM_NAME[0:-3].upper(),\
					datetime.datetime.now()\
				)\
			),\
		level=logging.DEBUG,\
		format=formatString
	)


def handlingCmdArguments():
	"""
	handlingCmdArguments
	--------------------

	This function configurates the ArgumentParser with the specific details of
	this programs.

	Does not take parameters.
	"""
	parser = argparse.ArgumentParser(\
		prog="{0} (v.{1}.{2}.{3})".format(PROGRAM_NAME,VERSION,MIN_VERSION,FIX_VERSION),\
		formatter_class=argparse.RawDescriptionHelpFormatter,\
		description=\
			'''\033[1m
================================================================================
  NGSphy
================================================================================
\033[0m
NGSphy is a Python open-source tool for the genome-wide simulation of NGS data
(read counts or Illumina reads) obtained from thousands of gene families evolving
under a common species tree, with multiple haploid and/or diploid individuals per
species, where sequencing coverage (depth) heterogeneity can vary among
individuals and loci, including off-target loci and phylogenetic decay effects.

For more information about usage and installation please go to the README file
or to the wiki page https://gihub.com/merlyescalona/ngsphy/wiki/
			''',\
		epilog="Developed by:\n{0}\n{1}\n\nVersion:\t{2}.{3}.{4} (Under development)\n{5}\n".format(AUTHOR,INSTITUTION, VERSION,MIN_VERSION,FIX_VERSION, LINE),\
		add_help=False
		)
	optionalGroup= parser.add_argument_group("{0}Optional arguments{1}".format("\033[1m","\033[0m"))
	optionalGroup.add_argument('-s','--settings',metavar='<settings_file_path>', type=str,\
	help='Path to the settings file.')
	optionalGroup.add_argument('-l','--log',metavar='<log_level>', type=str,\
		choices=LOG_LEVEL_CHOICES, default="INFO",\
		help='Specified level of log that will be shown through the standard output. Log will be stored in a separate file when level==DEBUG. Values:{0}. Default: {1}. '.format(LOG_LEVEL_CHOICES,LOG_LEVEL_CHOICES[1]))
	informationGroup= parser.add_argument_group("{0}Information arguments{1}".format("\033[1m","\033[0m"))
	informationGroup.add_argument('-v', '--version',\
		action='version',\
		version='{0}: Version {1}.{2}.{3}'.format(PROGRAM_NAME,VERSION,MIN_VERSION,FIX_VERSION),\
		help="Show program's version number and exit")
	informationGroup.add_argument('-h', '--help',\
		action='store_true',\
		help="Show this help message and exit")

	status=True; message=""
	try:
		tmpArgs = parser.parse_args()
		if (tmpArgs.help):
			parser.print_help()
		# Checks if the log level debug is chosen it will also print the debug file
		if tmpArgs.log==LOG_LEVEL_CHOICES[0]:
			APPLOGGER.setLevel(logging.DEBUG)
			createLogFile()
		if tmpArgs.log==LOG_LEVEL_CHOICES[1]: APPLOGGER.setLevel(logging.INFO)
		if tmpArgs.log==LOG_LEVEL_CHOICES[2]: APPLOGGER.setLevel(logging.WARNING)
		if tmpArgs.log==LOG_LEVEL_CHOICES[3]: APPLOGGER.setLevel(logging.ERROR)
	except:
		message="{0}\n{1}\n{2}\n".format(\
			"Something happened while parsing the arguments.",\
			"Please verify. Exiting.", LINE)
		APPLOGGER.error(message)
		parser.print_help()
		sys.exit(-1)
	if not tmpArgs.settings and not os.path.exists(os.path.abspath(os.path.join(os.getcwd(),"settings.txt"))):
		parser.print_help()
	return tmpArgs

###############################   MAIN   #######################################
def main():
	"""
	main()
	--------------------

	Entry point of the NGSphy program.
	"""
	try:
		cmdArgs = handlingCmdArguments()
		prog = ngsphy.NGSphy(cmdArgs)
		prog.run()
	except ngsphy.NGSphyExitException as ex:
		if ex.expression:
			APPLOGGER.info("NGSphy finished properly.")
			APPLOGGER.info("Elapsed time (ETA):\t{0}".format(ex.time))
			APPLOGGER.info("Ending at:\t{0}".format(datetime.datetime.now().strftime("%a, %b %d %Y. %I:%M:%S %p")))
			sys.exit()
		else:
			APPLOGGER.error(ex.message)
			APPLOGGER.error("Elapsed time (ETA):\t{0}".format(ex.time))
			APPLOGGER.error("Ending at:\t{0}".format(datetime.datetime.now().strftime("%a, %b %d %Y. %I:%M:%S %p")))
			sys.exit(-1)
	except KeyboardInterrupt:
		APPLOGGER.error("{0}{1}\nProgram has been interrupted.{2}\nPlease run again for the expected outcome.\n{3}\n".format("\033[91m","\033[1m","\033[0m",LINE))
		sys.exit(-1)

if __name__=="__main__":
	main()



from setuptools import setup, find_packages
def readme():
    with open('README.md') as f:
        return f.read()
long_description='NGSphy is a Python open-source tool for the genome-wide simulation of NGS data (read counts or Illumina reads) obtained from thousands of gene families evolving under a common species tree, with multiple haploid and/or diploid individuals per species, where sequencing coverage (depth) heterogeneity can vary among species, individuals and loci, including off-target or uncaptured loci.'

setup(name='ngsphy',\
    version='1.0.11',\
    description='phylogenomic simulation of NGS data ',\
    long_description=long_description,\
    url='https://github.com/merlyescalona/ngsphy',\
    download_url='https://github.com/merlyescalona/ngsphy/blob/master/dist/ngsphy-1.0.11.tar.gz',\
    author='Merly Escalona',\
    author_email='merlyescalona@uvigo.es',\
    license='GNU/GPL v3',\
    packages=['ngsphy'],\
    package_dir={'ngsphy': 'ngsphy'},\
    py_modules = [\
        'ngsphy.coverage', \
        'ngsphy.individual', \
        'ngsphy.loggingformatter', \
        'ngsphy.msatools', \
        'ngsphy.ngsphyexceptions', \
        'ngsphy.readcounts', \
        'ngsphy.reads', \
        'ngsphy.rerooter', \
        'ngsphy.sequence', \
        'ngsphy.settings'
    ],\
    install_requires=[\
        'argparse',\
        'ConfigParser',\
        'datetime',\
        'dendropy',\
        'logging',\
        'multiprocessing',\
        'numpy',\
        'setuptools',\
        'scipy'\
    ],
    package_data={
        'data':['data/*'],\
        'doc':['doc/*'],\
    },\
    classifiers=[\
        'Development Status :: 4 - Beta',\
        'Intended Audience :: Education',\
        'Intended Audience :: Science/Research',\
        'Intended Audience :: Developers',\
        'Topic :: Scientific/Engineering :: Bio-Informatics',\
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',\
        'Programming Language :: Python :: 2.7'\
    ],\
    keywords='biology phylogenomics next-generation sequencing coverage targeted-sequencing',\
    python_requires='~=2.7',\
    scripts=['scripts/ngsphy'],\
    entry_points={
        'console_scripts':[\
            'ngsphy = ngsphy.__main__:main'\
        ]\
    },\
    zip_safe=False\
  )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import copy,dendropy,glob,logging,os,sys, sqlite3
import numpy as np
import random as rnd
import settings as sp
from msatools import *
from select import select
from ngsphyexceptions import *
################################################################################
class IndividualAssignment:
	"""
	Class for the generation of individuals
	----------------------------------------------------------------------------
	Attributes:
	- appLogger: logger to store status of the process flow
	- settings: Settings object withh all the program parameters
	relation to the generated individuals
	- numReplicates: number of species trees.
	- numReplicateDigits: number of digits needed to represent numReplicates.
	- numLociPerReplicate: number of loci per species tree.
	- numLociPeReplicateDigits: number of digits needed to represent.
	- numIndividualsPerReplicate: number of individuals per species tree.
	- filteredST: identifier of the species trees that will be used.
	"""
	appLogger=None
	settings=None
	output=""
	numReplicates=0
	numReplicateDigits=0
	numLociPerReplicate=[]
	numLociPerReplicateDigits=[]
	numIndividualsPerReplicate=[]
	filteredReplicates=[]
	indelsPresence=False

	def __init__(self, settings):
		self.appLogger=logging.getLogger("ngsphy")
		self.appLogger.info("IndividualGenerator: Run started")
		self.settings=settings

	def checkArgs(self):
		"""
		Validation of the values needed for this process.
		- Input according origin mode
		"""
		self.appLogger.info("Checking arguments folder...")
		self.generateFolderStructure()
		message=""
		# need to know whether i'm working with simphy or indelible
		self.numReplicates=self.settings.numReplicates
		self.filteredReplicates=range(1,self.numReplicates+1)
		self.numReplicateDigits=len(str(self.numReplicates))
		self.numIndividualsPerReplicate=[0]*len(self.filteredReplicates)
		self.numLociPerReplicate=[0]*self.numReplicates
		self.numLociPerReplicateDigits=[0]*self.numReplicates
		########################################################################
		if self.settings.inputmode < 4 :
			# this is like this, because for this to work it is necessary that
			# sequences had been generated, and it is in the SequenceGenerator
			# that the numLociPerReplicate option is set up.
			status,message=self.checkPloidyTreeRelation()
			if not status: return status,message
			if not self.settings.parser.has_option("general","numlociperreplicate"):
				self=numLociPerReplicate=[1]
			else:
				self.numLociPerReplicate=[\
					self.settings.parser.getint("general","numlociperreplicate")]
			self.numLociPerReplicateDigits=[len(str(self.numLociPerReplicate[0]))]
		elif self.settings.inputmode==4:
			####################################################################
			# checking the replicate that are going to be used
			# checking if I'll used the filtered in case there's a possibility
			# that one or many sts do not match the ploidy and number of gene copies
			self.command = os.path.join(\
				self.settings.basepath,\
				"{0}.command".format(self.settings.projectName))
			self.params = os.path.join(\
				self.settings.basepath,\
				"{0}.params".format(self.settings.projectName))
			self.db = os.path.join(\
				self.settings.basepath,\
				"{0}.db".format(self.settings.projectName))

			self.numLociPerReplicate=self.getSimPhyNumLociPerSpeciesTree()
			self.numLociPerReplicateDigits=[len(str(a))for a in self.numLociPerReplicate]
			# check ploidy matches given data
			if self.settings.simphyFilter:
				self.appLogger.info("Filtering replicates...")
				self.filteredReplicates=self.filterReplicatesMatchingIndPerSpeciesAndPloidy(self.settings.ploidy)
				self.numLociPerReplicate=self.getSimPhyNumLociPerSpeciesTreeFiltered(self.filteredReplicates)
				self.numLociPerReplicateDigits=[len(str(a))for a in self.numLociPerReplicate]
			self.appLogger.info("Checking filtered replicates...")
			status,message=self.checkPloidySimPhyData()
			if not status: return status,message
 			# check that the species tree replicate folder have the correct data
			self.appLogger.info("Checking data within replicates...")
			gtperstOK,message=self.checkDataWithinReplicates()
			if (not gtperstOK): return gtperstOK,message
			self.settings.parser.set(\
				"general",\
				"numlociperreplicate",\
				",".join([str(a) for a in self.numLociPerReplicate]))
			self.printSimPhyConfiguration()
		else:
			return False, "{0}\n\t{1}\n\t{2}".format(\
				"Individual Assignment process."
				"Something is wrong with the input.",\
				"Please verify. Exiting."\
			)
		self.settings.parser.set("general","filtered_replicates",",".join([str(a) for a in self.filteredReplicates]))
		return True, message

	def generateFolderStructure(self):
		"""
		Generation of general folder structure needed for the individual generation
		"""
		self.appLogger.info("Creating folder structure for individual generation...")
		try:
			self.appLogger.info("Generating output folder {}".format(self.settings.individualsFolderPath))
			os.makedirs(self.settings.individualsFolderPath)
		except:
			self.appLogger.debug("Output folder exists ({0})".format(self.settings.outputFolderPath))
		try:
			self.appLogger.info("Generating output folder {}".format(self.settings.tablesFolderPath))
			os.makedirs(self.settings.tablesFolderPath)
		except:
			self.appLogger.debug("Output folder exists ({0})".format(self.settings.tablesFolderPath))


	def printSimPhyConfiguration(self):
		"""
		Print the configuration of the simphy project
		"""
		self.appLogger.debug(\
			"\n\tConfiguration...\n\t{0}\t{1}\n\t{2}\t{3}\n\t{4}\t{5}\n\t{6}\t{7}\n\t{8}\t{9}".format(\
				"SimPhy path:",\
				self.settings.basepath,\
				"SimPhy project name:",\
				self.settings.projectName,\
				"Output folder:",\
				self.settings.outputFolderPath,\
				"Dataset prefix(es) (INDELible):",\
				self.settings.simphyDataPrefix,\
				"Number of species trees replicates/folders:",\
				self.numReplicates\
			)
		)

	def getSimPhyNumLociPerSpeciesTree(self):
		"""
		Retrieves information of number of loci per species from the SimPhy
		database.
		------------------------------------------------------------------------
		Returns: a list with the number of loci per species tree replicate
		"""
		query="select N_Loci from Species_Trees"
		con = sqlite3.connect(self.db)
		res=con.execute(query).fetchall()
		con.close()
		res=[item for sublist in res for item in sublist]
		return res

	def getSimPhyNumLociPerSpeciesTreeFiltered(self, indices):
		"""
		Retrieves information of number of loci per species from the SimPhy
		database.
		------------------------------------------------------------------------
		Returns: a list with the number of loci per species tree replicate
		"""
		query="select N_Loci from Species_Trees where SID in ({0})".format(",".join([str(i) for i in indices]))
		con = sqlite3.connect(self.db)
		res=con.execute(query).fetchall()
		con.close()
		res=[item for sublist in res for item in sublist]
		return res

	def filterReplicatesMatchingIndPerSpeciesAndPloidy(self, ploidy):
		"""
		Identifies and filters the species tree replicates that  SimPhy
		database.
		------------------------------------------------------------------------
		Returns: a list with the number of loci per species tree replicate
		"""
		query="select SID from Species_Trees"
		if not (ploidy == 1):
			query="select SID from Species_Trees WHERE Ind_per_sp % {0} = 0".format(ploidy)
		con = sqlite3.connect(self.db)
		res=con.execute(query).fetchall()
		con.close()
		res=[item for sublist in res for item in sublist]
		return res

	def checkPloidySimPhyData(self):
		"""
		Checks whether the ploidy given matches the information in the
		SimPhy database
		------------------------------------------------------------------------
		Returns: a list with the number of loci per species tree replicate
		"""
		if not self.filteredReplicates == []:
			query="select Ind_per_sp from Species_Trees WHERE SID in ({})".format(",".join([str(i) for i in self.filteredReplicates]))
			con = sqlite3.connect(self.db)
			res=con.execute(query).fetchall()
			con.close()
			res=[item for sublist in res for item in sublist]
			status=True; message=""
			for item in res:
				if not item % self.settings.ploidy == 0:
					status=False
					message="{0}\n\t{1}\n\t{2}\n\t{3}".format(\
						"There has been a problem with the ploidy.",\
						"There is at least one species tree replicate, which number ",\
						"of individuals does not match the ploidy specified",\
						"Please verify. Exiting."
					)
		else:
			status=False
			message="{0}\n\t{1}\n\t{2}".format(\
				"There has been a problem with the ploidy.",\
				"No replicate satisfies the conditions.",\
				"Please verify. Exiting."
			)
		return status, message

	def checkPloidyTreeRelation(self):
		"""
		Checks whether the ploidy given matches the information given within
		the given tree.
		"""
		self.appLogger.debug("Checking ploidy - num tips relation")
		status=True
		message="Ploidy and number of gene copies per gene family match properly."
		tree=dendropy.Tree.get(\
			path=self.settings.geneTreeFile,\
		 	schema="newick",\
			preserve_underscores=True)
		leaves=[node.taxon.label for node in tree.leaf_node_iter()]
		leavesSplit=[ item.split("_") for item in leaves]
		leavesDict=dict()
		for tip in leavesSplit:
			geneFamily="_".join(tip[0:2])
			try:
				val=leavesDict[geneFamily]
				leavesDict[geneFamily]+=1
			except:
				leavesDict[geneFamily]=1
		for item in leavesDict:
			self.appLogger.debug("{0} - {1}".format(item,leavesDict[item]))
			if not leavesDict[item] % self.settings.ploidy == 0:
				status=False
				message+="\n\t{0}\n\t{1}\n\t{2}".format(\
				"INDELible control file - Something's wrong!",\
				"The number of gene copies within one of the gene families does not match the ploidy selected for this run.",\
				"Please verify. Exiting."\
				)

		return status, message

	def checkDataWithinReplicates(self):
		""""
		Checks the data files within the  replicates.
		Existence of fasta files.
		"""
		self.appLogger.info("Checking replicate data: ReplicateID - currentReplicate/numberOfWorkingReplicates [numberOfFiles]")
		# for indexREP in self.filteredReplicates:
		for index in range(0,len(self.filteredReplicates)):
			curReplicatePath=os.path.join(\
				self.settings.basepath,\
				"{0:0{1}d}".format(\
					self.filteredReplicates[index],\
					self.numReplicateDigits\
				)\
			)
			numFastaFiles=0;numGeneTrees=0
			# check composition of the current indexREP folder
			numFastaFiles=len(glob.glob("{0}/{1}_*_TRUE*".format(curReplicatePath, self.settings.simphyDataPrefix)))
			numGeneTrees=len(glob.glob("{0}/g_trees*.trees".format(curReplicatePath, self.settings.simphyDataPrefix)))
			self.numLociPerReplicate[index]=numFastaFiles
			self.appLogger.info("ReplicateID {0} - {1}/{2} [{3}]".format(\
				self.filteredReplicates[index],\
				index+1,\
			 	len(self.filteredReplicates),\
				numFastaFiles))
			self.numLociPerReplicateDigits[index]=len(str(numFastaFiles))
			if (numFastaFiles<1):
				# Do not have fasta files from the given replicate to work, I'll skip it.
				self.appLogger.warning(\
					"Replicate {0}({1}): It is not possible to do the mating for this replicate".format(self.filteredReplicates[index], curReplicatePath))
				return False, "Please verify. Exiting."
			if (numGeneTrees<1):
				return False,"Trying to mate sequences, but there are no gene tree files to back that up. Please, finish the SimPhy run and try again afterwards."
			self.appLogger.debug("numGeneTrees ({}) == numFastaFiles ({})? ".format(numGeneTrees,numFastaFiles))
			if not numGeneTrees == numFastaFiles:
				return False, "There are no sequences o there is a missmatch between the number of trees and the number of sequences in the folder."
		return True,"Got number of gene trees per species trees"

	def iteratingOverReplicates(self):
		"""
		Iterates over the species tree replicates and generates individuals
		according to the selected ploidy
		"""
		self.appLogger.debug("Ploidy: {0}".format(self.settings.ploidy))
		if (self.settings.ploidy==1):
			self.iterationHaploid()
		else:
			self.iterationPolyploid()
		self.settings.parser.set("general","numindividualsperreplicate",\
			",".join([str(a) for a in self.numIndividualsPerReplicate]))

	def iterationPolyploid(self):
		"""
		Iterates over the species tree replicates.
		Within each species tree, iterates over the gene trees, generates
		the "mating table" as well as the file with the individuals's sequences.
		"""
		self.appLogger.info("Generating individuals: replicateID - currentReplicate/numberOfWorkingReplicates [numLoci] (path)...")
		for index in range(0,len(self.filteredReplicates)):
			curReplicatePath=os.path.join(\
				self.settings.individualsFolderPath,\
				"REPLICATE_{0:0{1}d}".format(\
					self.filteredReplicates[index],\
					self.numReplicateDigits\
				)\
			)
			self.appLogger.info("ReplicateID {0} - {1}/{2} [{3}] ({4}) ".format(\
				self.filteredReplicates[index],\
				index+1,\
				len(self.filteredReplicates),\
				self.numLociPerReplicate[index],\
				curReplicatePath\
			))
			# iterating over the number of gts per st
			self.appLogger.debug("Generating individual table...")
			matingTable=self.generateMatingTable(self.filteredReplicates[index])
			self.numIndividualsPerReplicate[index]=len(matingTable)
			self.writeMatingTable(self.filteredReplicates[index],matingTable)
			for indexLOC in range(1,self.numLociPerReplicate[index]+1):
				# parsingMSA file
				fastapath=os.path.join(\
					self.settings.basepath,
					"{0:0{1}d}".format(self.filteredReplicates[index],self.numReplicateDigits),\
					"{0}_{1:0{2}d}.fasta".format(\
						self.settings.simphyDataPrefix,\
						indexLOC,\
						self.numLociPerReplicateDigits[index]
					)\
				)
				seqDict=parseMSAFile(fastapath)
				outputFolder=os.path.join(
					self.settings.individualsFolderPath,\
					"REPLICATE_{0:0{1}d}".format(\
						self.filteredReplicates[index],\
						self.numReplicateDigits),\
					"LOCUS_{0:0{1}d}".format(\
						indexLOC,\
						self.numLociPerReplicateDigits[index])\
				)
				try:
					self.appLogger.debug("Output folder: {0}".format(outputFolder))
					os.makedirs(outputFolder)
				except OSError as err:
					self.appLogger.warning("OS error: {0}".format(err))
					self.appLogger.debug("Folder {0} exists.".format(outputFolder))
				# generating and writing mating table
				self.mate(self.filteredReplicates[index],indexLOC,matingTable,seqDict)

	def iterationHaploid(self):
		"""
		Iterates over the species tree replicates.
		Within each species tree, iterates over the gene trees, generates
		the "relation table" as well as the file with the individuals's sequences.
		"""
		self.appLogger.info("Generating individuals: replicateID - currentReplicate/numberOfWorkingReplicates [numLoci] (path)...")
		# for indexREP in self.filteredReplicates:
		for index in range(0, len(self.filteredReplicates)):
			curReplicatePath=os.path.join(\
				self.settings.individualsFolderPath,
				"REPLICATE_{0:0{1}d}".format(\
					self.filteredReplicates[index],\
				 	self.numReplicateDigits)
			)
			self.appLogger.info("ReplicateID {0} - \t{1}/{2} [{3}] ({4}) ".format(\
				self.filteredReplicates[index],\
				index+1,\
				len(self.filteredReplicates),\
				self.numLociPerReplicate[index],\
				curReplicatePath\
			))
			# iterating over the number of gts per st
			individualTable=None
			self.appLogger.info("Generating individual table")
			individualTable=self.generateIndividualTable(self.filteredReplicates[index])
			self.numIndividualsPerReplicate[index]=len(individualTable)
			self.writeIndividualTable(self.filteredReplicates[index],individualTable)
			for indexLOC in range(1,self.numLociPerReplicate[index]+1):
				# parsingMSA file
				self.appLogger.debug("Using REP={0}, LOC={1}".format(self.filteredReplicates[index],indexLOC))
				fastapath=os.path.join(\
					self.settings.basepath,\
					"{0:0{1}d}".format(self.filteredReplicates[index],self.numReplicateDigits),\
					"{0}_{1:0{2}d}.fasta".format(\
						self.settings.simphyDataPrefix,\
						indexLOC,\
						self.numLociPerReplicateDigits[index]
					)\
				)
				seqDict=parseMSAFileWithDescriptions(fastapath)
				################################################################
				# if self.settings.inputmode==3:
				# 	# need to modify the sequence of the anchor-root
				# 	seqs=None
				# 	with open(self.settings.ancestralSequenceFilePath, 'r') as f:
				# 		seqs=[line.strip() for line in f if not line.startswith("#")]
				# 	seqDict[self.settings.anchorTipLabel]=seqs[0]
				################################################################
				outputFolder=os.path.join(\
					self.settings.individualsFolderPath,\
					"REPLICATE_{0:0{1}d}".format(\
						self.filteredReplicates[index], self.numReplicateDigits),\
					"LOCUS_{0:0{1}d}".format(\
						indexLOC,\
						self.numLociPerReplicateDigits[index])\
				)
				try:
					self.appLogger.debug("Output folder: {0}".format(outputFolder))
					os.makedirs(outputFolder)
				except OSError as err:
					self.appLogger.warning("OS error: {0}".format(err))
					self.appLogger.debug("Folder {0} exists.".format(outputFolder))
				# generating and writing mating table
				self.generateIndividuals(self.filteredReplicates[index],indexLOC,individualTable,seqDict)

	def generateIndividualTable(self,indexREP):
		"""
		Generates the table that stores the relationship between the original
		sequnce description and the final individual identifier.
		------------------------------------------------------------------------
		Parameters:
		- indexREP: identifier of the Species tree replicate
		Returns:
		- A table/matrix/list of lists with the relation between the orignal
		sequence and the corresponding individual identifier.
		"""
		# get first gt of the st to get the descriptions
		indexLOC=1;
		index=self.filteredReplicates.index(indexREP)
		self.appLogger.info("ReplicateID {0} - {1}/{2} [Locus {3}]".format(\
			indexREP,
			index,\
			len(self.filteredReplicates),\
			indexLOC\
			))
		fastapath=os.path.join(\
			self.settings.basepath,\
			"{0:0{1}d}".format(\
				self.filteredReplicates[index],\
				self.numReplicateDigits),\
			"{0}_{1:0{2}d}_TRUE.fasta".format(\
				self.settings.simphyDataPrefix,\
				indexLOC,\
				self.numLociPerReplicateDigits[index]\
			)\
		)
		descriptions=parseMSAFileWithDescriptions(fastapath).keys()
		descriptions.sort()
		table=[(item,descriptions[item]) for item in range(0,len(descriptions))]
		return table

	def generateIndividuals(self,indexREP,indexLOC,individualTable,seqDict):
		"""
		Once the table is generated and the corresponding file with the sequences
		of the specific gene tree is parsed, it is possible to generate the file
		with the sequences that correspond to an individual.
		------------------------------------------------------------------------
		Parameters:
		- indexREP: identifier of the Species tree replicate
		- indexLOC: identifier of the gene tree
		- individualTable: table with the relation between sequences and individuals.
		- seqDict: dictionary with the sequences parsed from the gene tree fasta file.

		Generates:
		- A set of files, as many as individuals were described in the table.
		"""
		index=self.filteredReplicates.index(indexREP)
		self.appLogger.debug(\
			"{0}/{1} - {2}".format(\
				indexLOC,\
				self.numLociPerReplicate[index],\
				self.numLociPerReplicateDigits[index]\
			)\
		)
		outputFolder=os.path.join(\
			self.settings.individualsFolderPath,\
			"REPLICATE_{0:0{1}d}".format(\
				self.filteredReplicates[index],\
				self.numReplicateDigits),\
			"LOCUS_{0:0{1}d}".format(\
				indexLOC,\
				self.numLociPerReplicateDigits[index])
		)
		for currentInd in range(0,len(individualTable)):
			# Extracting info from the dictionary
			indID=str(individualTable[currentInd][0])
			description=str(individualTable[currentInd][1])
			# Organizing strings
			# if description=="6_0_0":
				# print indID,outputFolder
			seq=seqDict[description]
			des=">{0}:{1:0{2}d}:{3:0{4}d}:{5}:{6}:{7}".format(
				self.settings.projectName,\
				self.filteredReplicates[index],\
				self.numReplicateDigits,\
				indexLOC,\
				self.numLociPerReplicateDigits[index],\
				self.settings.simphyDataPrefix,\
				indID,\
				description[1:len(description)]\
			)
			indFilename=os.path.join(\
				outputFolder,\
				"{0}_{1:0{2}d}_{3:0{4}d}_{5}_{6}.fasta".format(\
					self.settings.projectName,\
					self.filteredReplicates[index],\
					self.numReplicateDigits,\
					indexLOC,\
					self.numLociPerReplicateDigits[index],\
					self.settings.simphyDataPrefix,\
					indID\
				)\
			)
			indFile=open(indFilename, "w")
			indFile.write("{0}\n{1}\n".format(des,seq))
			indFile.close()

	def writeIndividualTable(self,indexREP,individualTable):
		"""
		Writes into a file the table (individualTable) with the relation between
		sequences and individuals identifiers for the specific indexREP species
		tree replicate
		------------------------------------------------------------------------
		Parameters:
		- indexREP: identifier of the Species tree replicate
		- individualTable: table with the relation between sequences and individuals.
		Generates:
		- File with the table for the indexREP species tree replicate
		"""
		# indexREP,SP,ind-tip1,ind-tip2
		index=self.filteredReplicates.index(indexREP)
		self.appLogger.debug("Writing indexes into file...")
		indexFilename=os.path.join(\
			self.settings.tablesFolderPath,\
			"{0}.{1:0{2}d}.individuals.csv".format(\
				self.settings.projectName,\
				self.filteredReplicates[index],\
				self.numReplicateDigits\
			)\
		)
		if not os.path.isfile(indexFilename):
			indexFile=open(indexFilename,"w")
			indexFile.write("repID,indID,spID,locID,geneID\n")
			indexFile.close()
		indexFile=open(indexFilename,"a")
		for indexRow in range(0,len(individualTable)):
			indID=individualTable[indexRow][0]
			seqDescription=individualTable[indexRow][1]
			speciesID=seqDescription.strip().split("_")[0]
			locusID=seqDescription.strip().split("_")[1]
			geneID=seqDescription.strip().split("_")[2]
			indexFile.write("{0:0{1}d},{2},{3},{4},{5}\n".format(\
				self.filteredReplicates[index],\
				self.numReplicateDigits,\
				indID,\
				speciesID,\
				locusID,\
				geneID))
		indexFile.close()


	# def generateMatingTableFromSequenceDescription(self,indexREP):
	def generateMatingTable(self,indexREP):
		"""
		Generates the "mating" table which stores the relation between the
		sequences that generates an individual.
		This table is generated from the sequence file of the first gene tree
		of the specific indexREP
		------------------------------------------------------------------------
		Parameters:
		- indexREP: identifier of the Species tree replicate
		Returns:
		- the mating table
		"""
		index=self.filteredReplicates.index(indexREP)
		self.appLogger.debug("Mating table")
		filename=os.path.join(\
			self.settings.basepath,\
			"{0:0{1}d}".format(\
				self.filteredReplicates[index],\
				self.numReplicateDigits),\
			"{0}_{1:0{2}d}.fasta".format(\
				self.settings.simphyDataPrefix,\
				1,\
				self.numLociPerReplicateDigits[index]))
		self.appLogger.debug("Reading file: {0}".format(filename))
		leaves=None
		with open(filename,"r") as f:
			leaves=[ line.strip()[1:] for line in f if line.startswith(">")]
		leavesSplit=[ item.split("_") for item in leaves if not item == self.settings.anchorTipLabel]
		leavesDict=dict()
		for tip in leavesSplit:
			geneFamily="_".join(tip[0:2])
			try:
				val=leavesDict[geneFamily]
				leavesDict[geneFamily]+=1
			except:
				leavesDict[geneFamily]=1
		# Till here i have information about the number of tips per gene family
		numLeaves=0
		for item in leavesDict:
			numLeaves+=leavesDict[item]
		mates=[]
		for geneFamily in leavesDict:
			if leavesDict[geneFamily]==1:
				sp=geneFamily.split("_")[0]
				lt=geneFamily.split("_")[1]
				pair=(self.filteredReplicates[index],sp,lt,0,0)
				mates+=[pair]
				self.appLogger.debug("Pair generated: {0}".format(pair))
			else:
				t=range(0,leavesDict[geneFamily])
				while not t==[]:
					p1=0;p2=0
					try:
						p1=t.pop(rnd.sample(range(0,len(t)),1)[0])
						p2=t.pop(rnd.sample(range(0,len(t)),1)[0])
					except Exception as e:
						break
					sp=geneFamily.split("_")[0]
					lt=geneFamily.split("_")[1]
					pair=(self.filteredReplicates[index],sp,lt,p1,p2)
					mates+=[pair]
					self.appLogger.debug("Pair generated: {0}".format(pair))
		return mates

	def writeMatingTable(self,indexREP,matingTable):
		"""
		Writes into a file the table (individualTable) with the relation between
		sequences and individuals identifiers for the specific indexREP species
		tree replicate
		------------------------------------------------------------------------
		Parameters:
		- indexREP: identifier of the Species tree replicate
		- matingTable: table with the relation between sequences and the generated
		individuals
		Generates:
		- File with the table for the indexREP species tree replicate
		"""
		index=self.filteredReplicates.index(indexREP)
		self.appLogger.debug("Writing indexes into file...")
		indexFilename=os.path.join(\
			self.settings.tablesFolderPath,\
			"{0}.{1:0{2}d}.individuals.csv".format(\
				self.settings.projectName,\
				self.filteredReplicates[index],\
				self.numReplicateDigits\
			)\
		)
		self.appLogger.debug(indexFilename)
		if not os.path.isfile(indexFilename):
			indexFile=open(indexFilename,"w")
			indexFile.write("repID,indID,spID,locID,mateID1,mateID2\n")
			indexFile.close()
		indexFile=open(indexFilename,"a")
		for indexRow in range(0,len(matingTable)):
			indID=indexRow
			speciesID=matingTable[indexRow][1]
			locusID=matingTable[indexRow][2]
			mateID1=matingTable[indexRow][3]
			mateID2=matingTable[indexRow][4]
			indexFile.write("{0:0{1}d},{2},{3},{4},{5},{6}\n".format(\
				self.filteredReplicates[index],\
				self.numReplicateDigits,\
				indID,\
				speciesID,\
				locusID,\
				mateID1,\
				mateID2))
		indexFile.close()

	def mate(self,indexREP,indexLOC,matingTable,sequenceDictionary):
		"""
		Once the table is generated and the corresponding file with the sequences
		of the specific gene tree is parsed, it is possible to generate the file
		with the sequences that correspond to an individual.
		------------------------------------------------------------------------
		Parameters:
		- indexREP: identifier of the Species tree replicate
		- indexLOC: identifier of the gene tree
		- matingTable: table with the relation between sequences and individuals.
		- seqDict: dictionary with the sequences parsed from the gene tree fasta file.
		Generates:
		- A set of files, as many as individuals were described in the table.
		"""
		index=self.filteredReplicates.index(indexREP)
		# Proper mating
		seqDict=copy.deepcopy(sequenceDictionary)
		species=seqDict.keys()
		numSeqs=0
		for key in species:
			subspecies=seqDict[key].keys()
			for subkey in subspecies:
				numSeqs+=len(seqDict[key][subkey])
		numInds=np.trunc(numSeqs/2)+1
		self.appLogger.debug("ReplicateID {0} -  {1}/{2} |  Locus {3:0{5}d}/{4:0{5}d}. Writing {6} individuals from {7} sequecens.".format(\
			self.filteredReplicates[index],\
			index+1,\
			len(self.filteredReplicates),\
			indexLOC,\
			self.numLociPerReplicate[index],\
			self.numLociPerReplicateDigits[index],\
			numSeqs,\
			numInds\
			)\
		)
		outputFolder=os.path.join(\
			self.settings.individualsFolderPath,\
			"REPLICATE_{0:0{1}d}".format(\
				self.filteredReplicates[index],\
				self.numReplicateDigits),\
			"LOCUS_{0:0{1}d}".format(\
				indexLOC,\
				self.numLociPerReplicateDigits[index])\
		)
		seq1="";des1="";seq2="";des2="";
		for currentInd in range(0,len(matingTable)):
			# Extracting info from the dictionary
			st=str(matingTable[currentInd][0])
			sp=str(matingTable[currentInd][1])
			lt=str(matingTable[currentInd][2])
			pair1=str(matingTable[currentInd][3])
			pair2=str(matingTable[currentInd][4])
			tag="{0}_{1}".format(sp,lt)
			# Organizing strings
			self.appLogger.debug("{0}|{1}-{2}".format(sp, pair1,pair2))
			seq1=seqDict[tag][pair1]["sequence"]
			seq2=seqDict[tag][pair2]["sequence"]
			shortDesc1=seqDict[tag][pair1]["description"]
			des1=">{0}:{1:0{2}d}:{3:0{4}d}:{5}:{6}:{7}_S1".format(\
				self.settings.projectName,\
				self.filteredReplicates[index],\
				self.numReplicateDigits,\
				indexLOC,\
				self.numLociPerReplicateDigits[index],\
				self.settings.simphyDataPrefix,\
				currentInd,shortDesc1\
			)
			shortDesc2=seqDict[tag][pair2]["description"]
			des2=">{0}:{1:0{2}d}:{3:0{4}d}:{5}:{6}:{7}_S2".format(\
				self.settings.projectName,\
				self.filteredReplicates[index],\
				self.numReplicateDigits,\
				indexLOC,\
				self.numLociPerReplicateDigits[index],\
				self.settings.simphyDataPrefix,\
				currentInd,\
				shortDesc2\
			)
			indFilename=os.path.join(\
				outputFolder,\
				"{0}_{1:0{2}d}_{3:0{4}d}_{5}_{6}.fasta".format(\
					self.settings.projectName,\
					self.filteredReplicates[index],\
					self.numReplicateDigits,\
					indexLOC,\
					self.numLociPerReplicateDigits[index],\
					self.settings.simphyDataPrefix,\
					currentInd\
				)\
			)
			indFile=open(indFilename, "w")
			indFile.write("{0}\n{1}\n{2}\n{3}\n".format(des1,seq1,des2,seq2))
			indFile.close()
			del seqDict[tag][pair1]
			if not sp== "0": del seqDict[tag][pair2]

		for tag in seqDict.keys():
			if not seqDict[tag]=={}:
				self.appLogger.warning("Number of individuals (sequences) generated per species is odd.")
				self.appLogger.warning("Sequence {0} from species {1} will not be paired.".format(\
					seqDict[tag].keys(), sp))
				for item in seqDict[tag].keys():
					del seqDict[sp][item]

		return None

	def checkFilesForIndels(self):
		"""
		Checks the FASTA files within the species tree replicates
		for indels. Since the RC process won't be ran if they appear in the
		dataset.
		------------------------------------------------------------------------
		Returns:
		- a boolean:
			TRUE if it does not have indel, meaning dataset is OK for processing.
			FALSE if it has indel, meaning dataset is INCORRECT for processing.

		"""
		self.appLogger.info("Checking for indels...")
		indelsList=[];self.indelsPresence=False;message=""
		endStatus=True
		try:
			for index in range(0,len(self.filteredReplicates)):
				self.appLogger.info("ReplicateID {0} - {1}/{2} [{3}] ".format(\
					self.filteredReplicates[index],\
					index+1,\
					len(self.filteredReplicates),\
					self.numLociPerReplicate[index]
				))
				genomicdata=""
				for indexLOC in range(1,self.numLociPerReplicate[index]+1):
					self.appLogger.debug("indexREP: {0}\tlocID: {1}".format(\
						self.filteredReplicates[index], indexLOC))
					fastapath=os.path.join(
						self.settings.basepath,\
						"{0:0{1}d}".format(self.filteredReplicates[index],self.numReplicateDigits),\
						"{0}_{1:0{2}d}_TRUE.fasta".format(\
							self.settings.simphyDataPrefix,\
							indexLOC,\
							self.numLociPerReplicateDigits[index]\
						)\
					)
					genomicdata=None
					with open(fastapath, "r") as fasta:
						genomicdata="".join([line.strip() for line in fasta if not line.startswith(">")])
					dashpos=int(genomicdata.find("-"))
					if dashpos >= 0: indelsList+=[True] #indel
					else: indelsList+=[False] # ok
			if sum(indelsList) > 0: self.indelsPresence=True
		except IOError as e:
			endStatus=False
			message="{}\t{} {} {}\n\t{}\n\t{}\n\t{}\n\t{}\n\n\t{}\n\t{}\n\t{}\n\t{}".format(\
				"[IOERROR]",\
				"File",\
				fastapath,\
				"does not exist",\
				"Input data should be in FASTA format. Please verify your dataset.",\
				"--------------------------------------------------------------------------------",\
				"If using input modes 1,2 or 3, make sure the control file is properly structured.",\
				"If using input mode 4, make sure the INDELible control file for your SimPhy simulations contains the settings lines:",\
				"[SETTINGS]",\
				"  [fastaextension] fasta",\
				"  [output] FASTA",\
				"--------------------------------------------------------------------------------"\
				)
		return endStatus, message



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import csv,logging,os,subprocess,sys,threading, time
import numpy as np
import random as rnd
from coverage import NGSPhyDistribution as ngsphydistro
import settings as sp

class RunningInfo(object):
	"""
	Class for the storage of running time information of ARTIllumina Class
	Separated to be able to be used by several threads.
	"""
	def __init__(self):
		# self.appLogger=logging.getLogger('sngsw')
		self.lock = threading.Lock()
		self.value = []

	def addLine(self, line):
		"""
		Adds a line to the running information estructure
		"""
		# self.appLogger.debug('Waiting for lock')
		self.lock.acquire()
		try:
			# self.appLogger.debug('Acquired lock')
			self.value += [line]
		finally:
			# self.appLogger.debug('Released lock')
			self.lock.release()

class ARTIllumina:
	"""
	Class for the generation of Illumina reads with the ART simulator.
	----------------------------------------------------------------------------
	Attributes:
	- CONSTANT: __SHORT_NAMES: all available coded-short parameters of the ART simulator
	- CONSTANT: __LONG_NAMES: all available  coded parameters of the ART simulator

	Technicallity. Parsing the simulator parameters from the settings file
	returns all of them into lower cases, in order to be able to compare them
	was necessarty to create these variable. Special cases are handled separate.

	- CONSTANT: __dLONG_NAMESHORT_NAMES: coded-short parameters of the ART simulator.
	- CONSTANT: __dLONG_NAMES: coded parameters of the ART simulator

	- params: final parameters used in the read simulation
	- commands: list of all the generated commands for the call of the ART simulator
	- numFiles: number of files that will be used for the simulation

	"""
	__SHORT_NAMES=["sf" ,"dp","ploidy","ofn","1","2","amp","c","d","ef","f","h","i",\
				"ir","ir2","dr","dr2","k","l","m","mp","M","nf","na",\
				"o","p","q","qU","qs","qL","qs2","rs","s","sam","sp","ss"]
	__LONG_NAMES=["simphy_folder","data_prefix","output_folder_name","ploidy","qprof1","qprof2",\
				"amplicon","rcount","id","errfree","fcov","help",\
				"in","insRate","insRate2","delRate","delRate2","maxIndel",\
				"len","mflen","matepair","cigarM","maskN","noALN","out",\
				"paired","quiet","maxQ","qShift","minQ","qShift2","rndSeed",\
				"sdev","samout","sepProf","seqSys"]
	__dLONG_NAMES={i.lower():i for i in __LONG_NAMES}
	__dSHORT_NAMES={i.lower():i for i in __SHORT_NAMES}

	params=[]
	commands=[]	 # Init of all the commands that will be generated
	numFiles=0

	individualsFileNameSuffix="individuals"

	def __init__(self,settings):
		self.appLogger=logging.getLogger("ngsphy")
		self.appLogger.info('NGS read simulation: ART run started.')
		self.settings=settings
		self.numReplicates=self.settings.numReplicates
		self.numReplicateDigits=len(str(self.numReplicates))
		self.filteredReplicates=[ int(numST) for numST in self.settings.parser.get("general", "filtered_replicates").strip().split(",")]
		self.numLociPerReplicate=[int(numST) for numST in self.settings.parser.get("general","numlociperreplicate").split(",")]
		self.numLociPerReplicateDigits=[ len(str(item)) for item in self.numLociPerReplicate]
		cc=self.settings.parser.get("general", "numindividualsperreplicate").strip().split(",")
		self.numIndividualsPerReplicate=[ int(item) for item in cc if not item == ""]
		self.numIndividualsPerReplicateDigits=[len(str(item )) for item in self.numIndividualsPerReplicate]

		dash=""; par=[]
		settingsParams=self.settings.parser.items("ngs-reads-art")
		for p in settingsParams:
			if (p[0] in self.__dSHORT_NAMES.keys()): dash="-"
			if (p[0] in self.__dLONG_NAMES.keys()): dash="--"
			# to be sure that i am getting the right parameter names
			if (dash=="-"):
				if (p[0]=="m" and ((par[1].lower() in ["true","false","on","off"]) or (par[1] in [0,1]))):
					par=["M"]
				elif (p[0]=="m"):
					par=["m"]
				else:
					par=[self.__dSHORT_NAMES[p[0]]]
			if (dash=="--"):
				par=[self.__dLONG_NAMES[p[0]]]

			par+=[p[1]]

			if((par[1].lower() in ["true","false","on","off"]) or (par[1] in [0,1])):
				self.params+=["{0}{1}".format(dash,par[0])]
			else:
				self.params+=["{0}{1}".format(dash,par[0]),par[1]]

		# generating specific folder structure
		self.settings.coverageFolderPath=os.path.join(self.settings.outputFolderPath,"coverage")
		self.settings.readsFolderPath=os.path.join(self.settings.outputFolderPath,"reads")
		self.settings.scriptsFolderPath=os.path.join(self.settings.outputFolderPath,"scripts")
		self.generateFolderStructure()
		if (self.settings.runningTimes): self.runningInfo=RunningInfo()


	def generateFolderStructure(self):
		"""
		Generation of basic output folder structure for this process
		"""
		self.appLogger.info("Creating folder structure for [ngs-reads-art]")
		try:
			os.makedirs(self.settings.readsFolderPath)
			self.appLogger.info("Generating output folder ({0})".format(self.settings.readsFolderPath))
		except:
			self.appLogger.debug("Output folder exists ({0})".format(self.settings.readsFolderPath))

		try:
			os.makedirs(self.settings.scriptsFolderPath)
			self.appLogger.info("Generating output folder ({0})".format(self.settings.scriptsFolderPath))
		except:
			self.appLogger.debug("Output folder exists ({0})".format(self.settings.scriptsFolderPath))


	def writeSGEScript(self):
		"""
		For the generation of scripts for cluster environments (Execution section,
		options run: SGE)
		This generates job script that will be launched in a SGE cluster.
		"""
		self.appLogger.debug("Start")
		jobfile=os.path.join(\
			self.settings.scriptsFolderPath,\
			"{0}.job.sge.sh".format(self.settings.projectName)\
		)

		j=open(jobfile,"w")
		seedfile=os.path.join(\
			self.settings.scriptsFolderPath,\
			"{0}.sh".format(self.settings.projectName)\
		)
		script="""#!/bin/bash
# SGE submission options
#$ -l num_proc=1		 # number of processors to use
#$ -l h_rt=00:10:00	  # Set 10 mins  - Average amount of time for up to 1000bp
#$ -t 1-{0}			  # Number of jobs/files that will be treated
#$ -N art.sims		   # A name for the job

command=$(awk 'NR==$SGE_TASK_ID{{print $1}}' {1})
$command
""".format(self.numFiles,seedfile)
		j.write(script)
		j.close()
		self.appLogger.info("SGE Job file written ({0})...".format(jobfile))

	def writeSLURMScript(self):
		"""
		For the generation of scripts for cluster environments (Execution section,
		options run: SLURM)
		This generates job script that will be launched in a SLURM cluster.
		"""
		self.appLogger.debug("Start")
		jobfile=os.path.join(\
			self.settings.scriptsFolderPath,\
			"{0}.job.slurm.sh".format(self.settings.projectName)
		)
		j=open(jobfile,"w")
		seedfile=os.path.join(\
			self.settings.scriptsFolderPath,\
			"{0}.sh".format(self.settings.projectName)\
		)

		script="""#!/bin/sh
#SBATCH -n 1
#SBATCH --cpus-per-task 1
#SBATCH -t 00:10:00
#SBATCH --mem 4G
#SBATCH --array=1-{0}

command=$(awk 'NR==$SLURM_ARRAY_TASK_ID{{print $1}}' {1})
$command


""".format(self.numFiles, seedfile)
		footer="".format()
		jobfile.write(script)
		jobfile.close()
		self.appLogger.info("SLURM Job file written ({0})...".format(jobfile))

	def retrieveCoverageMatrix(self, indexREP):
		"""
		Reads coverage matrix for a specific species tree identifier
		------------------------------------------------------------------------
		Parameters:
		- indexREP:specific species tree identifier
		Returns:
		- coverage Matrix
		"""
		index=self.filteredReplicates.index(indexREP)
		self.appLogger.debug("Retrieving coverage matrix")
		# coverage matrix per ST - row:indv - col:loci
		# each cov introduced as parameter is a NGSPhyDistribution
		coverageMatrixFilename=os.path.join(\
			self.settings.coverageFolderPath,\
			"{0}.{1:0{2}d}.coverage.csv".format(\
				self.settings.projectName,\
				self.filteredReplicates[index],\
				self.numReplicateDigits\
			)
		)
		coverageMatrix=np.zeros(shape=(self.numIndividualsPerReplicate[index], self.numLociPerReplicate[index]))
		firstLine=True; counter=0
		with open(coverageMatrixFilename, 'rb') as csvfile:
			coveragereader = csv.reader(csvfile, delimiter=',')
			for row in coveragereader:
				if firstLine:
					firstLine=False
				else:
					coverageMatrix[counter,]=[float(row[i]) for i in range(1,len(row))]
					counter+=1
				if not counter < self.numIndividualsPerReplicate[index]: break
		return coverageMatrix

	def getCommands(self):
		"""
		Generates command lines that will be called to generate the reads
		"""
		self.appLogger.debug("Start")
		for index in range(0, len(self.filteredReplicates)):
			csvfilename=os.path.join(\
				self.settings.tablesFolderPath,\
				"{0}.{1:0{2}d}.{3}.csv".format(\
					self.settings.projectName,\
					self.filteredReplicates[index],\
					self.numReplicateDigits,\
					self.individualsFileNameSuffix\
				)
			)
			with open(csvfilename) as csvfile:
				d = csv.DictReader(csvfile)
				self.matingDict = [row for row in d]
				csvfile.close()
				nInds=len(self.matingDict)
				nLoci=self.numLociPerReplicate[index]
				coverageMatrix=self.retrieveCoverageMatrix(self.filteredReplicates[index])
				if (self.settings.ploidy == 2):
					coverageMatrix=coverageMatrix*0.5
				for indexLOC in range(1,self.numLociPerReplicate[index]+1):
					for row in self.matingDict:
						# indexREP,indexLOC,indID,speciesID,mateID1,mateID2
						inputFile=os.path.join(\
							self.settings.individualsFolderPath,\
							"REPLICATE_{0:0{1}d}".format(\
								int(row['repID']),\
								self.numReplicateDigits
							),\
							"LOCUS_{0:0{1}d}".format(\
								indexLOC,\
								self.numLociPerReplicateDigits[index]\
							),\
							"{0}_{1:0{2}d}_{3:0{4}d}_{5}_{6}.fasta".format(\
								self.settings.projectName,\
								int(row['repID']),\
								self.numReplicateDigits,\
								indexLOC,\
								self.numLociPerReplicateDigits[index],
								self.settings.simphyDataPrefix,\
								int(row['indID'])\
							)\
						)
						# This means, from a multiple (2) sequence fasta file.
						outputFile=os.path.join(\
							self.settings.readsFolderPath,\
							"REPLICATE_{0:0{1}d}".format(\
								int(row['repID']),\
								self.numReplicateDigits
							),\
							"LOCUS_{0:0{1}d}".format(\
								indexLOC,\
								self.numLociPerReplicateDigits[index]\
							),\
							"{0}_{1:0{2}d}_{3:0{4}d}_{5}_{6}_R".format(\
								self.settings.projectName,\
								int(row['repID']),\
								self.numReplicateDigits,\
								indexLOC,\
								self.numLociPerReplicateDigits[index],
								self.settings.simphyDataPrefix,\
								int(row['indID'])\
							)\
						)
						coverage=coverageMatrix[int(row['indID'])][indexLOC-1]
						coverageParam=""
						if self.settings.coveragemodeART:
							coverageParam="--rcount"
						else:
							coverageParam="--fcov"
						# Call to ART
						callParams=["art_illumina"]+self.params+[coverageParam,str(coverage),"--in", inputFile,"--out",outputFile]
						# self.params+=["--in ",inputFile,"--out",outputFile]
						# print(callParams)
						self.commands+=[[row['repID'],indexLOC,row['indID'],inputFile, outputFile]+callParams]

		self.appLogger.info("Commands have been generated...")

	def writeBashScript(self):
		"""
		Generates a bash script with all the command lines used to generate
		the reads (ART commands)
		"""
		self.appLogger.debug("Start")
		bashfile=os.path.join(\
			self.settings.scriptsFolderPath,\
			"{0}.sh".format(self.settings.projectName)\
		)
		j=open(bashfile,"w")
		for item in self.commands:
			c=item[5:len(item)]
			j.write(" ".join(c))
			j.write("\n")
		j.close()

	def commandLauncher(self, command):
		"""
		Launches the execution of the third-party app for the simulation
		of  NGS reads.
		"""
		self.appLogger.debug("Start")
		ngsMessage="";proc=""
		try:
			proc = subprocess.check_output(command[5:],stderr=subprocess.STDOUT)
			cpuTime = [line for line in proc.split('\n') if "CPU" in line][0].split(":")[1]
			seed = [line for line in proc.split('\n') if "seed" in line][0].split(":")[1]
			# line=[(command[0:3],cpuTime,seed, command[4])]
			line=command[0:3]+[cpuTime,seed,command[4]]
			# ngsMessage="Command: {0} - Finished succesfully.".format(" ".join(command[5:]))
		except subprocess.CalledProcessError as error:
			ngsMessage="{}".format(error.output)+\
			"\n\n------------------------------------------------------------------------"+\
			"\n\nFor more information about this error please run the 'art' command separately.\n"+\
			"art_illumina command used:\n==========================\n"+\
			"{}\n\n".format(" ".join(command[5:]))
			self.appLogger.error(ngsMessage)
			line=command[0:3]+["-","-",command[4]]
			raise RuntimeError("\nART execution error. Please verify. Exciting.")
		if (self.settings.runningTimes): self.runningInfo.addLine(line)


	def run(self):
		"""
		Process flow for the generation of NGS reads
		"""
		status=True
		environment=self.settings.executionMode
		# Generating commands
		try:
			self.getCommands()
			self.generateFolderStructureNGS()
			if environment==1:
				message="ART run has finished succesfully."
				if (self.settings.runART):
					curr=0
					self.appLogger.info("Running...")
					while (curr < len(self.commands)):
						progress=(curr*100)/len(self.commands)
						sys.stdout.write("Progress {0:02.1f} %\r".format(progress))
						sys.stdout.flush()
						# print("Progress {0:02.1f} %\r".format(prog))
						command=self.commands[curr]
						t = threading.Thread(target=self.commandLauncher(command))
						t.start()
						curr=curr+1
						while (threading.activeCount()-1)==self.settings.numThreads:
							time.sleep(0.1)
					if (self.settings.runningTimes): self.writeRunningInfoIntoFile()
				else:
					self.appLogger.info("Environment BASH. Writing scripts")
					self.writeBashScript()
			if (environment==2):
				self.writeBashScript()
				self.appLogger.info("Environment SGE. Writing scripts")
				self.writeSeedFile()
			if (environment==3):
				self.appLogger.info("Environment SLURM. Writing scripts")
				self.writeBashScript()
				self.writeSLURMScript()
		except ValueError as verror:
			status=False
			message="\t{0}\n\t{1}\n\t{2}".format(\
				"Distribution parameter error:",\
				verror,\
				"Please verify. Exciting."\
			)
		except RuntimeError as rte:
			status=False
			message="\t{0}\n\t{1}\n\t{2}".format(\
				"Execution error.",\
				rte,\
				"Please verify. Exciting."\
			)

		return status, message

	def generateFolderStructureNGS(self):
		"""
		Generation of detail folder structure used for this process
		"""
		# iterating over commands to create folders
		folders=set([])
		for command in self.commands:
			infile=os.path.dirname(command[4])
			folders.add(infile)
		for item in folders:
			try:
				os.makedirs(item)
			except:
				self.appLogger.debug("Output folder exists ({0})".format(item))

	def writeRunningInfoIntoFile(self):
		"""
		Writes running time information into a file
		"""
		outputFile=os.path.join(self.settings.outputFolderPath,"{0}.info".format(self.settings.projectName))
		f=open(outputFile,"w")
		f.write("repID,locID,indID,inputFile,cpuTime,seed,outputFilePrefix\n")
		for item in self.runningInfo.value:
			f.write(
				str(item[0])+","+\
				str(item[1])+","+\
				str(item[2])+","+\
				str(item[3])+","+\
				str(item[4])+","+\
				item[5]+"\n"
			)
		f.close()
		self.appLogger.info("File with timings of the ART run can be find on: {0}".format(outputFile))
